from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List


class InferenceAdapter(metaclass=ABCMeta):
    """
    FilesystemAdapter interface
    """

    @abstractmethod
    def health(self) -> bool:
        """
        Determine if model is running.
        Arguments:
        Returns:
            True if the file existed
        """

    @abstractmethod
    def inference(self, ort_inputs: Dict[str, Any], ort_out_names: List[str]) -> List[Any]:
        """
        Determine if a directory exists.
        Arguments:
            ort_inputs: The dictionary of input
            ort_out_names: List of name of nodes wanna get value
        Returns:
            List of value of nodes
        """



"""
Module flysystem
"""

from typing import final

from typing_extensions import Self


class InferenceException(Exception):
    """
    Base exception class for AI inference client package
    """


@final
class UnableToInference(InferenceException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        self._model_name = ""
        self._model_version = ""
        self._reason = ""

    @final
    def model_name(self) -> str:
        return self._model_name

    @final
    def model_version(self) -> str:
        return self._model_version

    @final
    def reason(self) -> str:
        return self._reason

    @classmethod
    def with_model(cls, model_name: str, model_version: str, reason: str = "") -> Self:
        msg = f"Unable to inference with model: {model_name}-{model_version}. {reason}".rstrip()
        this = cls(msg)
        this._model_name = model_name
        this._model_version = model_version
        this._reason = reason
        return this



from abc import ABCMeta
from typing import Any, Dict

from .adapters import InferenceAdapter


class Inference(metaclass=ABCMeta):
    """
    Inference Client
    """

    def __init__(
        self,
        adapter: InferenceAdapter,
        config: Dict[str, Any] = None,
    ):
        self.adapter = adapter
        self.config = config

    def health(self) -> bool:
        return self.adapter.health()

    def inference(self, ort_inputs, ort_out_names):
        return self.adapter.inference(ort_inputs, ort_out_names)



import sys
import uuid

from typing import Any, Dict, List

from tritonclient.grpc import InferInput, InferRequestedOutput
from tritonclient.utils import np_to_triton_dtype

from ..adapters import InferenceAdapter
from ..error import UnableToInference
from .tritonclient_grpc_custom import InferenceServerClientCustom


class TritonInferenceAdapter(InferenceAdapter):
    def __init__(
        self,
        triton_server: str,
        model_name: str,
        version: str = "1",
        client_timeout: int = 30,
        ssl: bool = False,
        options: Dict[str, Any] = None,
    ) -> None:
        if options is None:
            options = [
                ("grpc.max_send_message_length", 512 * 1024 * 1024),
                ("grpc.max_receive_message_length", 512 * 1024 * 1024),
            ]

        self.grpc_client = InferenceServerClientCustom(url=triton_server, ssl=ssl, channel_args=options)
        if not (
            self.grpc_client.is_server_live(client_timeout=client_timeout)
            and self.grpc_client.is_server_ready(client_timeout=client_timeout)
        ):
            sys.exit("Triton server is not live")

        self.model_name = model_name
        self.version = version
        self.client_timeout = client_timeout

    def health(self) -> bool:
        if self.grpc_client.is_model_ready(self.model_name, self.version, client_timeout=self.client_timeout):
            return True
        return False

    def inference(self, ort_inputs: Dict[str, Any], ort_out_names: List[str]) -> List[Any]:
        inputs = []
        for input_name, input_data in ort_inputs.items():
            infer_input = InferInput(
                name=input_name, shape=input_data.shape, datatype=np_to_triton_dtype(input_data.dtype)
            )
            infer_input.set_data_from_numpy(input_data)
            inputs.append(infer_input)

        outputs = [InferRequestedOutput(output_name) for output_name in ort_out_names]

        try:
            res = self.grpc_client.infer(
                model_name=self.model_name,
                inputs=inputs,
                model_version=self.version,
                outputs=outputs,
                request_id=uuid.uuid4().__str__(),
                client_timeout=self.client_timeout,
            )
        except Exception as ex:
            raise UnableToInference.with_model(self.model_name, self.version, str(ex))

        if len(ort_out_names) == 1:
            return [res.as_numpy(ort_out_names[0])]
        return [res.as_numpy(out) for out in ort_out_names]



import abc
import time

from random import randint
from typing import Optional, Tuple

import grpc

from tritonclient.grpc import InferenceServerClient, service_pb2_grpc


class SleepingPolicy(abc.ABC):
    @abc.abstractmethod
    def sleep(self, try_i: int):
        """
        How long to sleep in milliseconds.
        :param try_i: the number of retry (starting from zero)
        """
        assert try_i >= 0


class ExponentialBackoff(SleepingPolicy):
    def __init__(self, *, init_backoff_ms: int, max_backoff_ms: int, multiplier: int):
        self.init_backoff = randint(0, init_backoff_ms)
        self.max_backoff = max_backoff_ms
        self.multiplier = multiplier

    def sleep(self, try_i: int):
        sleep_range = min(self.init_backoff * self.multiplier**try_i, self.max_backoff)
        sleep_ms = randint(0, sleep_range)
        time.sleep(sleep_ms / 1000)


class RetryOnRpcErrorClientInterceptor(grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor):
    def __init__(
        self,
        *,
        max_attempts: int,
        sleeping_policy: SleepingPolicy,
        status_for_retry: Optional[Tuple[grpc.StatusCode]] = None,
    ):
        self.max_attempts = max_attempts
        self.sleeping_policy = sleeping_policy
        self.status_for_retry = status_for_retry

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):
        for try_i in range(self.max_attempts):
            response = continuation(client_call_details, request_or_iterator)
            if isinstance(response, grpc.RpcError):
                # Return if it was last attempt
                if try_i == (self.max_attempts - 1):
                    return response

                # If status code is not in retryable status codes
                if self.status_for_retry and response.code() not in self.status_for_retry:
                    return response

                self.sleeping_policy.sleep(try_i)
            else:
                return response

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return self._intercept_call(continuation, client_call_details, request_iterator)


class InferenceServerClientCustom(InferenceServerClient):
    def __init__(
        self,
        url,
        verbose=False,
        ssl=False,
        root_certificates=None,
        private_key=None,
        certificate_chain=None,
        creds=None,
        keepalive_options=None,
        channel_args=None,
    ):
        super().__init__(
            url,
            verbose,
            ssl,
            root_certificates,
            private_key,
            certificate_chain,
            creds,
            keepalive_options,
            channel_args,
        )
        interceptors = (
            RetryOnRpcErrorClientInterceptor(
                max_attempts=4,
                sleeping_policy=ExponentialBackoff(init_backoff_ms=100, max_backoff_ms=1600, multiplier=2),
                status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
            ),
        )
        self._client_stub = service_pb2_grpc.GRPCInferenceServiceStub(
            grpc.intercept_channel(self._channel, *interceptors),
        )



import sys

from os.path import join
from typing import Any, Dict, List

import onnxruntime

from ..adapters import InferenceAdapter


class OnnxInferenceAdapter(InferenceAdapter):
    def __init__(self, model_name: str, version: str = "1", limit_mem_gpu: int = -1, logger_level: int = 3) -> None:
        onnxruntime.set_default_logger_severity(logger_level)
        providers = ["CPUExecutionProvider"]
        if onnxruntime.get_available_providers() == providers or limit_mem_gpu <= 0:
            # Run with CPU
            self.device_type = "cpu"
        elif limit_mem_gpu > 0:
            # Run with GPU
            self.device_type = "cuda"
            providers = [
                (
                    "CUDAExecutionProvider",
                    {
                        "device_id": 0,
                        "arena_extend_strategy": "kNextPowerOfTwo",
                        "gpu_mem_limit": limit_mem_gpu * 1024 * 1024 * 1024,
                        "cudnn_conv_algo_search": "EXHAUSTIVE",
                        "do_copy_in_default_stream": True,
                    },
                ),
                "CPUExecutionProvider",
            ]
        else:
            sys.exit("Not Support Device")
        if not model_name.endswith("model.onnx"):
            model_name = join(model_name, version, "model.onnx")
        self.ort_session = onnxruntime.InferenceSession(model_name, providers=providers)

    def health(self) -> bool:
        if self.ort_session is None:
            return False
        return True

    def inference(self, ort_inputs: Dict[str, Any], ort_out_names: List[str]) -> List[Any]:
        # IOBinding
        io_binding = self.ort_session.io_binding()
        for ort_input_name, ort_input in ort_inputs.items():
            io_binding.bind_cpu_input(ort_input_name, ort_input)
        for ort_out_name in ort_out_names:
            io_binding.bind_output(ort_out_name, self.device_type)

        self.ort_session.run_with_iobinding(io_binding)
        ort_outs = io_binding.copy_outputs_to_cpu()
        io_binding.clear_binding_inputs()
        io_binding.clear_binding_outputs()
        del io_binding

        return ort_outs



from os import getenv
from os.path import dirname, join

import cv2
import numpy as np
import pytest

from infer_client.adapters.onnx import OnnxInferenceAdapter
from infer_client.adapters.triton import TritonInferenceAdapter

IMG_SIZE = 224
infer_onnx_obj = OnnxInferenceAdapter(
    model_name=join(dirname(__file__), "resources/test_classify"), version="1", limit_mem_gpu=-1
)
infer_triton_obj = TritonInferenceAdapter(
    triton_server=getenv("TRITON_SERVER"), model_name="test_classify", version="1", ssl=True
)


def preprocess(img, img_size):
    x = cv2.resize(img, img_size, interpolation=cv2.INTER_LINEAR)
    x = (x / 255.0).astype(np.float32)
    return x.transpose(2, 0, 1)[None, ...]


def postprocess(output):
    e_x = np.exp(output)
    return np.argmax(e_x / e_x.sum())


@pytest.mark.parametrize(
    "path,expected",
    (
        (join(dirname(__file__), "resources/dog.jpg"), 264),
        (join(dirname(__file__), "resources/cat.jpg"), 285),
    ),
)
def test_inference(path: str, expected: str):
    img = cv2.imread(path)
    preprocessed_img = preprocess(img, (IMG_SIZE, IMG_SIZE))

    res_onnx = infer_onnx_obj.inference({"x": preprocessed_img}, ["400"])
    res_triton = infer_triton_obj.inference({"x": preprocessed_img}, ["400"])

    assert np.allclose(res_onnx, res_triton, rtol=1.0e-4)

    if not res_onnx:
        return None
    pred = res_onnx[0]
    output = postprocess(pred)
    assert output.item() == expected
